//===================================================================================
// Microsoft patterns & practices
// Composite Application Guidance for Windows Presentation Foundation and Silverlight
//===================================================================================
// Copyright (c) Microsoft Corporation.  All rights reserved.
// THIS CODE AND INFORMATION IS PROVIDED "AS IS" WITHOUT WARRANTY
// OF ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING BUT NOT
// LIMITED TO THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
// FITNESS FOR A PARTICULAR PURPOSE.
//===================================================================================
// The example companies, organizations, products, domain names,
// e-mail addresses, logos, people, places, and events depicted
// herein are fictitious.  No association with any real company,
// organization, product, domain name, email address, logo, person,
// places, or events is intended or should be inferred.
//===================================================================================
namespace Microsoft.Practices.Composite.Modularity
{
    using System;
    using System.Collections.Generic;
    using System.Linq;

    /// <summary>
    /// Implements a <see cref="IModuleEnumerator"/> that enumerates through module metadata defined explicitly.
    /// </summary>
    public class StaticModuleEnumerator : IModuleEnumerator
    {
        private readonly List<ModuleInfo> startupModules = new List<ModuleInfo>();
        private readonly List<ModuleInfo> demandModules = new List<ModuleInfo>();


        /// <summary>
        /// Gets a list of metadata information of the modules.
        /// </summary>
        /// <returns>An array of <see cref="ModuleInfo"/>.</returns>
        public ModuleInfo[] GetModules()
        {
            return this.startupModules.Union(this.demandModules).ToArray();
        }

        /// <summary>
        /// Gets a list of metadata information of the modules that should be loaded at startup.
        /// </summary>
        /// <returns>An array of <see cref="ModuleInfo"/>.</returns>
        public ModuleInfo[] GetStartupLoadedModules()
        {
            return this.startupModules.ToArray();
        }

        /// <summary>
        /// Gets the metadata information of a module by its name.
        /// </summary>
        /// <param name="moduleName">The module's name.</param>
        /// <returns>A <see cref="ModuleInfo"/> associated with the <paramref name="moduleName"/> parameter.</returns>
        public ModuleInfo GetModule(string moduleName)
        {
            return this.demandModules.Union(this.startupModules).FirstOrDefault(moduleInfo => moduleInfo.ModuleName == moduleName);
        }

        /// <summary>
        /// Registers a module with the <see cref="StaticModuleEnumerator" />.
        /// </summary>
        /// <param name="moduleType">The module type. This class should implement <see cref="IModule"/>.</param>
        /// <param name="dependsOn">The names of the modules that this module depends on, if any.</param>
        /// <returns>The same instance of <see cref="StaticModuleEnumerator"/>.</returns>
        /// <remarks>The module name will be the Name of the type specified in <paramref name="moduleType"/>.</remarks>
        public StaticModuleEnumerator AddModule(Type moduleType, params string[] dependsOn)
        {
            return this.AddModule(moduleType, true, dependsOn);
        }

        /// <summary>
        /// Registers a module with the <see cref="StaticModuleEnumerator" />.
        /// </summary>
        /// <param name="moduleType">The module type. This class should implement <see cref="IModule"/>.</param>
        /// <param name="startupLoaded"><see langword="true"/> to load this module at startup, <see langword="false"/> otherwise</param>
        /// <param name="dependsOn">The names of the modules that this module depends on, if any.</param>
        /// <returns>The same instance of <see cref="StaticModuleEnumerator"/>.</returns>
        /// <remarks>The module name will be the Name of the type specified in <paramref name="moduleType"/>.</remarks>
        public StaticModuleEnumerator AddModule(Type moduleType, bool startupLoaded, params string[] dependsOn)
        {
            ModuleInfo moduleInfo = new ModuleInfo(moduleType.Name, moduleType.AssemblyQualifiedName);
            moduleInfo.DependsOn.AddRange(dependsOn);
            if (startupLoaded)
            {
                this.startupModules.Add(moduleInfo);
            }
            else
            {
                this.demandModules.Add(moduleInfo);
            }

            return this;
        }
    }
}