//===============================================================================
// Microsoft patterns & practices
// Mobile Client Software Factory - July 2006
//===============================================================================
// Copyright  Microsoft Corporation.  All rights reserved.
// THIS CODE AND INFORMATION IS PROVIDED "AS IS" WITHOUT WARRANTY
// OF ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING BUT NOT
// LIMITED TO THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
// FITNESS FOR A PARTICULAR PURPOSE.
//===============================================================================
// The example companies, organizations, products, domain names,
// e-mail addresses, logos, people, places, and events depicted
// herein are fictitious.  No association with any real company,
// organization, product, domain name, email address, logo, person,
// places, or events is intended or should be inferred.
//===============================================================================

using System;

namespace Microsoft.Practices.Mobile.ObjectBuilder
{
	/// <summary>
	/// Implementation of <see cref="IBuilderStrategy"/> which will notify an object about
	/// the completion of a <see cref="IBuilder{T}.BuildUp"/> operation, or start of a
	/// <see cref="IBuilder{T}.TearDown"/> operation.
	/// </summary>
	/// <remarks>
	/// This strategy checks the object that is passing through the builder chain to see if it
	/// implements IBuilderAware and if it does, it will call <see cref="IBuilderAware.OnBuiltUp"/>
	/// and <see cref="IBuilderAware.OnTearingDown"/>. This strategy is meant to be used from the
	/// <see cref="BuilderStage.PostInitialization"/> stage.
	/// </remarks>
	public class BuilderAwareStrategy : BuilderStrategy
	{
		/// <summary>
		/// See <see cref="IBuilderStrategy.BuildUp"/> for more information.
		/// </summary>
		public override object BuildUp(IBuilderContext context, Type typeToBuild, object existing, string idToBuild)
		{
			IBuilderAware awareObject = existing as IBuilderAware;
			
			if (awareObject != null)
			{
				//TraceBuildUp(context, t, id, Properties.Resources.CallingOnBuiltUp);
				awareObject.OnBuiltUp(idToBuild);
			}
			return base.BuildUp(context, typeToBuild, existing, idToBuild);
		}

		/// <summary>
		/// See <see cref="IBuilderStrategy.TearDown"/> for more information.
		/// </summary>
		public override object TearDown(IBuilderContext context, object item)
		{
			IBuilderAware awareObject = item as IBuilderAware;
			
			if (awareObject != null)
			{
			
				//TraceTearDown(context, item, Properties.Resources.CallingOnTearingDown);
				awareObject.OnTearingDown();
			}
			return base.TearDown(context, item);
		}
	}
}
