//===============================================================================
// Microsoft patterns & practices
// Mobile Client Software Factory - July 2006
//===============================================================================
// Copyright  Microsoft Corporation.  All rights reserved.
// THIS CODE AND INFORMATION IS PROVIDED "AS IS" WITHOUT WARRANTY
// OF ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING BUT NOT
// LIMITED TO THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
// FITNESS FOR A PARTICULAR PURPOSE.
//===============================================================================
// The example companies, organizations, products, domain names,
// e-mail addresses, logos, people, places, and events depicted
// herein are fictitious.  No association with any real company,
// organization, product, domain name, email address, logo, person,
// places, or events is intended or should be inferred.
//===============================================================================

using System;
using Microsoft.Practices.Mobile.ObjectBuilder;
using System.Reflection;
using System.Globalization;

namespace Microsoft.Practices.Mobile.ObjectBuilder
{
	/// <summary>
	/// Strongly-typed method invocation information that allows clients to pass method handlers for the 
	/// actual method invocation and parameter retrieval for invoking the method.
	/// </summary>
	/// <devdoc>
	/// ANY code that uses the mobile version of this class CANNOT be run on the desktop. See 
	/// http://lab.msdn.microsoft.com/ProductFeedback/viewFeedback.aspx?feedbackId=FDBK50657 and 
	/// GetMethodImplementationFlags commented-out method below.
	/// </devdoc>
	public class StrongTypedMethodCallInfo : IMethodCallInfo
	{
		MethodInvokeHandler methodInvoke;
		MethodParametersHandler getParameters;
		Type[] parameterTypes;

		/// <summary>
		/// Initializes the policy with the method invocation delegate, parameter retrieval delegate and optional array of parameter types.
		/// </summary>
		/// <param name="methodHandler">Method that will be called with the parameters retrieved by <paramref name="getParameters"/>, if any.</param>
		/// <param name="getParameters">Handler for retrieving the parameters to pass to the method handler. 
		/// Can be <see langword="null" /> if no <paramref name="parameterTypes"/> are provided.</param>
		/// <param name="parameterTypes">Optional list of parameter types used by the method and parameter retrieval handlers.</param>
		public StrongTypedMethodCallInfo(MethodInvokeHandler methodHandler, MethodParametersHandler getParameters, params Type[] parameterTypes)
		{
			Guard.ArgumentNotNull(methodHandler, "constructor");
			Guard.ArgumentNotNull(getParameters, "getParameters");

			this.methodInvoke = methodHandler;
			this.getParameters = getParameters;
			this.parameterTypes = parameterTypes;
		}

		/// <summary>
		/// Implements <see cref="IMethodCallInfo.SelectMethod"/>, returning a custom <see cref="MethodInfo"/> that 
		/// will use the method handler implementation received in the constructor.
		/// </summary>
		public MethodInfo SelectMethod(IBuilderContext context, Type type, string id)
		{
			return new StrongTypedMethod(methodInvoke, parameterTypes);
		}

		/// <summary>
		/// Implements the <see cref="IMethodCallInfo.GetParameters"/> calling the parameter retrieval handler received 
		/// in the constructor.
		/// </summary>
		public object[] GetParameters(IBuilderContext context, Type type, string id, MethodInfo method)
		{
			return getParameters(context, id);
		}

		class StrongTypedMethod : MethodInfo
		{
			MethodInvokeHandler methodInvoke;
			Type[] parameterTypes;

			public StrongTypedMethod(MethodInvokeHandler methodInvoke, Type[] parameterTypes)
			{
				this.methodInvoke = methodInvoke;
				this.parameterTypes = parameterTypes;
			}

			public override object Invoke(object obj, BindingFlags invokeAttr, Binder binder, object[] parameters, CultureInfo culture)
			{
				return methodInvoke(obj, parameters);
			}

			public override ParameterInfo[] GetParameters()
			{
				if (parameterTypes.Length == 0)
				{
					return new ParameterInfo[0];
				}

				ParameterInfo[] infos = new ParameterInfo[parameterTypes.Length];
				for (int i = 0; i < infos.Length; i++)
				{
					infos[i] = new StrongTypedParameter(parameterTypes[i]);
				}

				return infos;
			}

			#region NotSupported Members

			/// <summary>
			/// OB does not use the return value, so we do not implement this functionality.
			/// </summary>
			public override Type ReturnType
			{
				get { throw new NotImplementedException(); }
			}

			public override MethodInfo GetBaseDefinition()
			{
				throw new NotImplementedException();
			}

			public override ICustomAttributeProvider ReturnTypeCustomAttributes
			{
				get { throw new NotImplementedException(); }
			}

#if !PocketPC
			
			// This override is required on the desktop CF, but it doesn't show up 
			// as an overridable member, so it doesn't compile. As a consequence, 
			// there's no way to run tests that use this class on the desktop.
			public override MethodImplAttributes GetMethodImplementationFlags()
			{
				throw new NotImplementedException();
			}
#endif

			public override MethodAttributes Attributes
			{
				get { throw new NotImplementedException(); }
			}

			public override RuntimeMethodHandle MethodHandle
			{
				get { throw new NotImplementedException(); }
			}

			public override Type DeclaringType
			{
				get { throw new NotImplementedException(); }
			}

			public override object[] GetCustomAttributes(Type attributeType, bool inherit)
			{
				throw new NotImplementedException();
			}

			public override object[] GetCustomAttributes(bool inherit)
			{
				throw new NotImplementedException();
			}

			public override bool IsDefined(Type attributeType, bool inherit)
			{
				throw new NotImplementedException();
			}

			public override string Name
			{
				get { throw new NotImplementedException(); }
			}

			public override Type ReflectedType
			{
				get { throw new NotImplementedException(); }
			}

			#endregion

			class StrongTypedParameter : ParameterInfo
			{
				public StrongTypedParameter(Type parameterType)
				{
					base.ClassImpl = parameterType;
				}
			}
		}
	}
}
